package it.inaf.oats.ia2.tap.tsm.rest;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flipkart.zjsonpatch.JsonDiff;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import it.inaf.oats.ia2.tap.tsm.datalayer.ConnectionManager;
import it.inaf.oats.ia2.tap.tsm.datalayer.UpdateManager;
import it.inaf.oats.ia2.tap.tsm.model.Credentials;
import it.inaf.oats.ia2.tap.tsm.model.DbEntity;
import it.inaf.oats.ia2.tap.tsm.model.IDbEntitiesContainer;
import it.inaf.oats.ia2.tap.tsm.model.Status;
import it.inaf.oats.ia2.tap.tsm.model.TapSchema;
import java.io.IOException;
import javax.ws.rs.core.Context;
import javax.ws.rs.Consumes;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.ws.rs.FormParam;
import javax.ws.rs.POST;

/**
 * REST Web Service
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
@Path("tapschema")
public class TapSchemaResource {

    private static final String PAGE_STATE = "pageState";

    // containers/entity names
    private static final String TAP_SCHEMA = "tapSchema";
    private static final String SCHEMA = "schema";
    private static final String TABLE = "table";
    private static final String COLUMN = "column";

    private static final Gson STANDARD_GSON = new GsonBuilder().create();
    private static final Gson NULLABLE_GSON = new GsonBuilder().serializeNulls().create();
    private static final ObjectMapper mapper = new ObjectMapper();

    @Context
    private HttpServletRequest request;

    private final Credentials credentials = new Credentials("localhost", 3306, "root", "root");

    private TapSchema getTapSchema() {
        return getPageState().getTapSchema();
    }

    private SchemaPageState getPageState() {
        SchemaPageState pageState = null;
        HttpSession session = request.getSession();
        if (session.getAttribute(PAGE_STATE) == null) {

            Connection connection = null;
            try {
                connection = credentials.getConnection();
                pageState = new SchemaPageState(new TapSchema(connection, "vlkb_voinfo_schema"));
                session.setAttribute(PAGE_STATE, pageState);
            } catch (SQLException e) {
                e.printStackTrace(System.err);
            } finally {
                try {
                    if (connection != null) {
                        connection.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace(System.err);
                }
            }
        } else {
            pageState = (SchemaPageState) session.getAttribute(PAGE_STATE);
        }
        return pageState;
    }

    @POST
    @Path("model")
    @Produces(MediaType.APPLICATION_JSON)
    public String getPageModelJson() {
        PageModel pi = new PageModel(getTapSchema());
        return NULLABLE_GSON.toJson(pi);
    }

    private String getPageModelPatches() throws IOException {
        SchemaPageState pageState = getPageState();

        String oldModelJson = NULLABLE_GSON.toJson(pageState.getPreviousModel());

        PageModel newModel = new PageModel(pageState.getTapSchema());
        pageState.setPreviousModel(newModel);
        String newModelJson = NULLABLE_GSON.toJson(newModel);

        JsonNode jn1 = mapper.readTree(oldModelJson);
        JsonNode jn2 = mapper.readTree(newModelJson);

        JsonNode patches = JsonDiff.asJson(jn1, jn2);

        return patches.toString();
    }

    private IDbEntitiesContainer getEntitiesContainer(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(TAP_SCHEMA)) {
            return tapSchema;
        }
        if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedSchema();
        }
        if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        }
        return null;
    }

    private DbEntity getDbEntity(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedSchema();
        }
        if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        }
        if (name.equals(COLUMN)) {
            return tapSchema.getSelectedColumn();
        }
        return null;
    }

    @POST
    @Path("select")
    public String select(@FormParam("containerName") String containerName, @FormParam("entityName") String entityName) throws IOException {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        container.select(entityName);
        return getPageModelPatches();
    }

    class RemoveColumnResult {

        PageModel pageModelUpdates;
        boolean toRemove;
    }

    @POST
    @Path("remove")
    public String remove(@FormParam("containerName") String containerName, @FormParam("entityName") String entityName) throws IOException {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        container.removeEntity(entityName);
        return getPageModelPatches();
    }

    class Addable {

        boolean selected;
        String name;

        Addable(String name) {
            this.name = name;
        }
    }

    @POST
    @Path("addables")
    public String getAddables(@FormParam("containerName") String containerName) {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        ArrayList<Addable> addables = new ArrayList<Addable>();
        for (String name : container.getAddables()) {
            addables.add(new Addable(name));
        }
        return STANDARD_GSON.toJson(addables);
    }

    class AddParams {

        String containerName;
        List<String> entities;
    }

    @POST
    @Path("update")
    public String update() throws SQLException {
        TapSchema tapSchema = getTapSchema();
        (new UpdateManager(credentials, tapSchema)).update();
        return "updated";
    }

    @POST
    @Path("add")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String add(String params) throws SQLException, IOException {
        final AddParams addParams = STANDARD_GSON.fromJson(params, AddParams.class);

        (new ConnectionManager(credentials) {

            @Override
            public void use(Connection connection) throws SQLException {
                getEntitiesContainer(addParams.containerName).loadEntities(connection, addParams.entities);
            }
        }).connect();

        return getPageModelPatches();
    }

    class EditParams {

        String entity;
        String key;
        String value;
    }

    @POST
    @Path("edit")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String edit(String params) throws SQLException {
        EditParams editParams = STANDARD_GSON.fromJson(params, EditParams.class);
        return getDbEntity(editParams.entity).updateValue(editParams.key, editParams.value) + "";
    }

    @POST
    @Path("undoremove")
    @Produces(MediaType.APPLICATION_JSON)
    public String undoremove(@FormParam("containerName") String containerName) throws SQLException, IOException {
        getEntitiesContainer(containerName).getSelected().setStatus(Status.SELECTED_SAVED);
        return getPageModelPatches();
    }

    @POST
    @Path("searchUCD")
    @Produces(MediaType.APPLICATION_JSON)
    public String searchUCD(@FormParam("description") String description) throws Exception {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();
        if (description == null || description.equals("")) {
            dialog.setDefault();
        } else {
            dialog.search(description);
        }

        return NULLABLE_GSON.toJson(dialog);
    }

    @POST
    @Path("selectUCD")
    public String selectUCD(@FormParam("index") String indexStr) {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();

        if (indexStr == null || indexStr.equals("")) {
            dialog.selectUCD();
        } else {
            dialog.selectUCD(Integer.parseInt(indexStr));
        }

        return dialog.getSelectedUCD();
    }

    @POST
    @Path("explainUCD")
    public String explainUCD(@FormParam("index") String indexStr) throws Exception {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();

        if (indexStr == null || indexStr.equals("")) {
            return dialog.explain();
        }

        return dialog.explain(Integer.parseInt(indexStr));
    }
}
